
from shexer.utils.uri import remove_corners, add_corners, RDF_TYPE
from shexer.model.node_selector import NodeSelectorNoSparql, NodeSelectorSparql
from rdflib.plugins import sparql
import re

_QUOTES = ["'", '"']
_WHITES_REGEX = re.compile(" +")

_FOCUS_LOWER = "focus"
_WILDCARD = "_"

_FOCUS_VARIABLE = "?f"
_WILDCARD_VARIABLE = "?x"


class NodeSelectorParser(object):

    def __init__(self, prefix_namespaces_dict, sgraph):
        self._prefix_namespace_dict = prefix_namespaces_dict
        self._sgraph = sgraph
        # self._endpoint_url = endpoint_url

    def parse_node_selector(self, raw_selector):
        raw_selector = raw_selector.strip()
        if raw_selector.startswith("<"):
            return self._parse_unprefixed_node_selector(raw_selector)
        elif raw_selector.startswith("{"):
            return self._parse_focus_expression(raw_selector)
        elif raw_selector.startswith("SPARQL"):
            return self._parse_sparql_expression(raw_selector)
        else:
            return self._parse_prefixed_node_selector(raw_selector)

    def _parse_unprefixed_node_selector(self, raw_selector):
        return NodeSelectorNoSparql(raw_selector=raw_selector,
                                    target_node=remove_corners(raw_selector),
                                    sgraph=self._sgraph)

    def _parse_prefixed_node_selector(self, raw_selector):
        for a_prefix in self._prefix_namespace_dict:
            if raw_selector.startswith(a_prefix + ":"):
                return NodeSelectorNoSparql(raw_selector=raw_selector,
                                            target_node=self._unprefix_uri(prefix=a_prefix,
                                                                           uri=raw_selector),
                                            sgraph=self._sgraph)

    def _parse_focus_expression(self, raw_selector):
        if raw_selector[0] != "{" or raw_selector[-1] != "}":
            raise ValueError("The following node selector is not surrounded by {}: " * raw_selector)
        raw_string = raw_selector[1:-1].strip()
        raw_string = _WHITES_REGEX.sub(" ", raw_string)
        pieces = raw_string.split(" ")
        if len(pieces) != 3:
            self._focus_node_error(raw_selector)
        subject_for_query, focus_count = self._parse_subj_obj_focus_expression(pieces[0], 0)
        predicate_for_query = self._parse_uri_focus_expression(pieces[1])
        object_for_query, focus_count = self._parse_subj_obj_focus_expression(pieces[2], focus_count)

        if focus_count != 1:
            raise ValueError("The node selector must have exactly one FOCUS")

        query = self._turn_focus_exp_tokens_into_query(subject_for_query, predicate_for_query, object_for_query)
        return NodeSelectorSparql(raw_selector=raw_selector,
                                  sparql_query_selector=query,
                                  id_variable_query=self._parse_variable_in_single_variable_query(query),
                                  sgraph=self._sgraph)

    def _turn_focus_exp_tokens_into_query(self, subj, pred, obj):
        return self._namespaces_to_string() + "SELECT " + _FOCUS_VARIABLE + " WHERE {" + subj + " " + pred + " " + obj + " . } "
        # return sparql.prepareQuery(string_query, initNs=self._prefix_namespace_dict)

    def _parse_subj_obj_focus_expression(self, token, focus_count):
        if token.lower() == _FOCUS_LOWER:
            return _FOCUS_VARIABLE, focus_count + 1
        elif token == _WILDCARD:
            return _WILDCARD_VARIABLE, focus_count
        else:
            return self._parse_uri_focus_expression(token), focus_count

    def _parse_uri_focus_expression(self, token):
        if token == "a":
            return add_corners(RDF_TYPE)
        elif token.endswith(">"):
            if token.startswith("<"):
                return token
        else:
            for a_prefix in self._prefix_namespace_dict:
                if token.startswith(a_prefix + ":"):
                    return add_corners(self._unprefix_uri(prefix=a_prefix,
                                              uri=token))
        raise ValueError("URI not well formed or with an unknown prefix: " + token)

    def _unprefix_uri(self, prefix, uri):
        return uri.replace(prefix + ":", self._prefix_namespace_dict[prefix])

    def _parse_sparql_expression(self, raw_selector):
        raw_string = raw_selector.replace("SPARQL", "")
        raw_string = raw_string.strip()
        if raw_string[0] in _QUOTES and raw_string[-1] in _QUOTES:
            try:
                return self._parse_single_variable_select_query(raw_string[1:-1])
            except BaseException as e:
                raise ValueError("The SPARQL query of the next node selector is not well formed: " \
                        + raw_selector + ". Cause: " + str(e))
        raise ValueError("The SPARQL query of the next node selector is not surrounded by quotes: " + raw_selector)

    def _parse_single_variable_select_query(self, string_query):
        # Is the query well-formed? If not, the next sentence raises error
        sparql.prepareQuery(string_query, initNs=self._prefix_namespace_dict)
        # Is it a select query?
        if "select" not in string_query[:string_query.find("{")].lower():
            raise ValueError("The SPARQL query is not a SELECT query")
        # Does it have a single variable
        if string_query[:string_query.find("{")].count("?") != 1:
            raise ValueError("The SPARQL query must have a single variable")

        variable_id = self._parse_variable_in_single_variable_query(string_query)

        return NodeSelectorSparql(raw_selector=string_query,
                                  sparql_query_selector=self._namespaces_to_string() + string_query,
                                  id_variable_query=variable_id,
                                  sgraph=self._sgraph)

    def _parse_variable_in_single_variable_query(self, string_query):
        index_first_char_var_name = string_query.find('?') + 1
        index_last_char_var_name = string_query[index_first_char_var_name:].find(" ") + index_first_char_var_name
        return string_query[index_first_char_var_name:index_last_char_var_name]

    def _namespaces_to_string(self):
        namespaces = ""
        for prefix, uri in self._prefix_namespace_dict.items():
            namespaces += "PREFIX " + prefix + ": <" + uri + ">\n"
        return namespaces

    @staticmethod
    def _focus_node_error(raw_selector):
        raise ValueError("This focus node expression cant be parsed: " + raw_selector)



class ZipFileLineReader(object):
    def __init__(self, zip_archive, zip_target):
        self._zip_archive = zip_archive
        self._zip_target = zip_target

    def read_lines(self):
        with self._zip_archive.open(self._zip_target, "r") as in_stream:
            for a_line in in_stream:
                yield a_line.decode("utf-8")




import sys
from datetime import datetime

def _curr_time():
    return datetime.now().strftime("%d/%m/%Y %H:%M:%S -- ")


# def log_to_error(msg, source=None, target_log=sys.stderr):
#     # pass
#     print(msg + ". Source: " + (source if source is not None else "Not specified"), file=target_log)


def log_msg(verbose, msg, err=True):
    if verbose:
        print(_curr_time() + msg, flush=True, file=None if not err else sys.stderr)


import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from rdflib import Graph

from shexer.consts import TURTLE

_BASE_DIR = BASE_FILES + "general" + pth.sep

class TestGraphFileInput(unittest.TestCase):

    def test_parsing_file(self):
        a_g = Graph()
        a_g.parse(G1, format="turtle")

        shaper = Shaper(target_classes=["http://xmlns.com/foaf/0.1/Person",
                                        "http://xmlns.com/foaf/0.1/Document"],
                        rdflib_graph=a_g,
                        namespaces_dict=default_namespaces(),
                        all_classes_mode=False,
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))

    def test_all_classes_mode(self):
        a_g = Graph()
        a_g.parse(G1, format="turtle")

        shaper = Shaper(all_classes_mode=True,
                        rdflib_graph=a_g,
                        namespaces_dict=default_namespaces(),
                        input_format=TURTLE,
                        disable_comments=True)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "g1_all_classes_no_comments.shex",
                                                      str_target=str_result))




import unittest
from shexer.shaper import Shaper
from test.const import G1, BASE_FILES, default_namespaces, G1_ALL_CLASSES_NO_COMMENTS
from test.t_utils import file_vs_str_tunned_comparison
import os.path as pth
from shexer.consts import TURTLE, SHAPES_DEFAULT_NAMESPACE

_BASE_DIR = BASE_FILES + "shapes_namespace" + pth.sep # We just need something with another instantiation property


class TestShapesNamespaces(unittest.TestCase):

    def test_default_namespace(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            shapes_namespace=SHAPES_DEFAULT_NAMESPACE)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=G1_ALL_CLASSES_NO_COMMENTS,
                                                      str_target=str_result))

    def test_different_namespace(self):
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=default_namespaces(),
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            shapes_namespace="http://weso.es/DIfferentShapes#")
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "different_namespace.shex",
                                                      str_target=str_result))

    def test_empty_prefix_used(self):
        namespaces = default_namespaces()
        namespaces["http://unuseful.but.yet/here/"] = ""
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=namespaces,
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            shapes_namespace=SHAPES_DEFAULT_NAMESPACE)
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "empty_prefix_used.shex",
                                                      str_target=str_result))

    def test_empty_prefix_used_and_no_default(self):
        namespaces = default_namespaces()
        namespaces["http://unuseful.but.yet/here/"] = ""
        shaper = Shaper(
            graph_file_input=G1,
            namespaces_dict=namespaces,
            all_classes_mode=True,
            input_format=TURTLE,
            disable_comments=True,
            shapes_namespace="http://weso.es/DIfferentShapes#")
        str_result = shaper.shex_graph(string_output=True)
        self.assertTrue(file_vs_str_tunned_comparison(file_path=_BASE_DIR + "empty_prefix_used_and_no_def.shex",
                                                      str_target=str_result))



class Literal(object):

    def __init__(self, content, elem_type):
        self._content = content
        self._elem_type = elem_type

    def __str__(self):
        return self._content

    @property
    def elem_type(self):
        return self._elem_type






from shexer.core.instances.pconsts import _S, _P, _O
from shexer.core.instances.annotators.base_annotator import BaseAnnotator

class AnnotatorTrackingInstances(BaseAnnotator):

    def __init__(self, instance_tracker):
        super().__init__(instance_tracker)

    def annotate_triple(self, a_triple):
        if self._instance_tracker.is_subclass_property(a_triple[_P]):
            self._annotate_subclass(a_triple)
        else:
            super().annotate_triple(a_triple)

    def is_relevant_triple(self, a_triple):
        if a_triple[_P] == self._subclass_property:
            return True
        else:
            return super().annotate_triple(a_triple)

    def annotation_post_parsing(self):
        for a_key_class in self._instances_dict:
            self._classes_considered_in_htree.add(a_key_class)
        iri_node = self._htree.iri_node
        for a_key_class in self._classes_considered_in_htree:
            a_class_node = self._get_appropiate_iri_node_and_add_to_htree_if_needed(a_key_class)
            if not a_class_node.has_parents():
                a_class_node.add_parent(iri_node)

    def _get_appropiate_iri_node_and_add_to_htree_if_needed(self, str_iri):
        return self._htree.get_node_of_element(str_iri) if self._htree.contains_element(str_iri) \
            else self._htree.create_node_IRI(str_iri)

    def _annotate_subclass(self, a_triple):
        str_s = str(a_triple[_S])
        str_o = str(a_triple[_O])

        subj_node = self._get_appropiate_iri_node_and_add_to_htree_if_needed(str_s)
        obj_node = self._get_appropiate_iri_node_and_add_to_htree_if_needed(str_o)
        subj_node.add_parent(obj_node)

        self._classes_considered_in_htree.add(str_s)
        self._classes_considered_in_htree.add(str_o)



from shexer.core.shexing.class_shexer import ClassShexer
from shexer.consts import RDF_TYPE, SHAPES_DEFAULT_NAMESPACE, RATIO_INSTANCES


def get_class_shexer(class_counts,
                     class_profile_dict,
                     remove_empty_shapes=True,
                     original_target_classes=None,
                     original_shape_map=None,
                     discard_useless_constraints_with_positive_closure=True,
                     keep_less_specific=True,
                     all_compliant_mode=True,
                     instantiation_property=RDF_TYPE,
                     disable_or_statements=True,
                     disable_comments=False,
                     namespaces_dict=None,
                     allow_opt_cardinality=True,
                     disable_exact_cardinality=False,
                     shapes_namespace=SHAPES_DEFAULT_NAMESPACE,
                     inverse_paths=False,
                     decimals=-1,
                     instances_report_mode=RATIO_INSTANCES,
                     detect_minimal_iri=False,
                     class_min_iris=None,
                     allow_redundant_or=False):

    return ClassShexer(
        class_counts_dict=class_counts,
        class_profile_dict=class_profile_dict,
        class_profile_json_file=None,
        remove_empty_shapes=remove_empty_shapes,
        original_target_classes=original_target_classes,
        original_shape_map=original_shape_map,
        discard_useless_constraints_with_positive_closure=discard_useless_constraints_with_positive_closure,
        keep_less_specific=keep_less_specific,
        all_compliant_mode=all_compliant_mode,
        instantiation_property=instantiation_property,
        disable_or_statements=disable_or_statements,
        disable_comments=disable_comments,
        namespaces_dict=namespaces_dict,
        allow_opt_cardinality=allow_opt_cardinality,
        disable_exact_cardinality=disable_exact_cardinality,
        shapes_namespace=shapes_namespace,
        inverse_paths=inverse_paths,
        instances_report_mode=instances_report_mode,
        decimals=decimals,
        detect_minimal_iri=detect_minimal_iri,
        class_min_iris_dict=class_min_iris,
        allow_redundant_or=allow_redundant_or
    )



import os.path as pth

BASE_FILES = pth.join(pth.dirname(pth.normpath(__file__)), "t_files" + pth.sep)
BASE_FILES_GENERAL = BASE_FILES + "general" + pth.sep

G1 = BASE_FILES + "t_graph_1.ttl"
G1_NT = BASE_FILES + "t_graph_1.nt"
G1_TSVO_SPO = BASE_FILES + "t_graph_1.tsv"
G1_JSON_LD = BASE_FILES + "t_graph_1.json"
G1_XML = BASE_FILES + "t_graph_1.xml"
G1_TTL_WITH_BASE = BASE_FILES + "t_graph_1_base.ttl"
G1_TTL_WITH_ABSOLUTES = BASE_FILES + "t_graph_1_absolutes.ttl"
G1_TTL_WITH_USELESS_BNODE = BASE_FILES + "t_graph_1_bnode.ttl"
G1_TTL_WITH_SCAPED_QUOTES = BASE_FILES + "t_graph_1_scaped_quotes.ttl"

G1_N3 = BASE_FILES + "t_graph_1.n3"

G1_ALL_CLASSES_NO_COMMENTS = BASE_FILES_GENERAL + "g1_all_classes_no_comments.shex"


# PREFIX xml: <http://www.w3.org/XML/1998/namespace/>
# PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
# PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
# PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
# PREFIX foaf: <http://xmlns.com/foaf/0.1/>

# NAMESPACES_WITH_FOAF_AND_EX = {"http://example.org/" : "ex",
#                                "http://www.w3.org/XML/1998/namespace/" : "xml",
#                                "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf",
#                                "http://www.w3.org/2000/01/rdf-schema#" : "rdfs",
#                                "http://www.w3.org/2001/XMLSchema#": "xsd",
#                                "http://xmlns.com/foaf/0.1/": "foaf"
#                                }

def default_namespaces():
    return {"http://example.org/": "ex",
            "http://www.w3.org/XML/1998/namespace/": "xml",
            "http://www.w3.org/1999/02/22-rdf-syntax-ns#": "rdf",
            "http://www.w3.org/2000/01/rdf-schema#": "rdfs",
            "http://www.w3.org/2001/XMLSchema#": "xsd",
            "http://xmlns.com/foaf/0.1/": "foaf"
            }



from shexer.consts import RDF_TYPE

class SGraph(object):

    def __init__(self):
        pass

    def query_single_variable(self, str_query, variable_id):
        """
        It receives an SPARQL query with a single variable and returns a list with the nodes matching that query

        :param str_query:
        :param variable_id:
        :return: list
        """
        raise NotImplementedError()

    def yield_p_o_triples_of_an_s(self, target_node):
        """
        Here it expects unprefixed URIs. So there is no stage of namespaces management to build a query.

        :param target_node:
        :return:
        """
        raise NotImplementedError()

    def yield_s_p_triples_of_an_o(self, target_node):
        """
        Here it expects unprefixed URIs. So there is no stage of namespaces management to build a query.
        :param target_node:
        :return:
        """
        raise NotImplementedError()

    def yield_class_triples_of_an_s(self, target_node, instantiation_property):
        """
        Here it expects unprefixed URIs. So there is no stage of namespaces management to build a query.

        :param target_node:
        :param instantiation_property:
        :return:
        """
        raise NotImplementedError()

    def yield_s_p_triples_of_target_nodes(self, target_nodes, depth, classes_at_last_level=True,
                                          instantiation_property=RDF_TYPE, already_visited=None,
                                          strict_syntax_with_uri_corners=True
                                          ):
        """
        If it is provided, the param already_visited can be modified during the execution of this method.
        The set already_visited can be used to avoid repetition of triples calling this methodd repeatedly
        for different node selectors in a shape map.

        :param target_nodes:
        :param depth:
        :param classes_at_last_level:
        :param instantiation_property:
        :param already_visited:
        :param strict_syntax_with_uri_corners:
        :return:
        """
        current_already_visited = set() if already_visited is None else already_visited
        list_of_current_target_nodes = target_nodes
        new_target_nodes = []
        while depth > 0:
            for a_node in list_of_current_target_nodes:
                if a_node not in current_already_visited:
                    current_already_visited.add(a_node)
                    for a_triple in self.yield_s_p_triples_of_an_o(a_node):
                        yield a_triple
                        if self._is_an_unprefixed_iri(an_iri=a_triple[0],
                                                      strict_syntax_with_uri_corners=strict_syntax_with_uri_corners):
                            new_target_nodes.append(a_triple[0])
            depth -= 1
            list_of_current_target_nodes = new_target_nodes
            new_target_nodes = []
            if depth == 0 and classes_at_last_level:
                for a_node in list_of_current_target_nodes:
                    if a_node not in current_already_visited:
                        for a_triple in self.yield_class_triples_of_an_s(target_node=a_node,
                                                                         instantiation_property=instantiation_property):
                            yield a_triple

    def yield_p_o_triples_of_target_nodes(self, target_nodes, depth, classes_at_last_level=True,
                                          instantiation_property=RDF_TYPE, already_visited=None,
                                          strict_syntax_with_uri_corners=True):
        """
        If it is provided, the param already_visited can be modified during the execution of this method.
        The set already_visited can be used to avoid repetition of triples calling this methodd repeatedly
        for different node selectors in a shape map.

        :param target_nodes:
        :param depth:
        :param classes_at_last_level:
        :param instantiation_property:
        :param already_visited:
        :param strict_syntax_with_uri_corners:
        :return:
        """

        current_already_visited = set() if already_visited is None else already_visited
        list_of_current_target_nodes = target_nodes
        new_target_nodes = []
        while depth > 0:
            for a_node in list_of_current_target_nodes:
                if a_node not in current_already_visited:
                    current_already_visited.add(a_node)
                    for a_triple in self.yield_p_o_triples_of_an_s(a_node):
                        yield a_triple
                        if self._is_an_unprefixed_iri(an_iri=a_triple[2],
                                                      strict_syntax_with_uri_corners=strict_syntax_with_uri_corners):
                            new_target_nodes.append(a_triple[2])
            depth -= 1
            list_of_current_target_nodes = new_target_nodes
            new_target_nodes = []
            if depth == 0 and classes_at_last_level:
                for a_node in list_of_current_target_nodes:
                    if a_node not in current_already_visited:
                        for a_triple in self.yield_class_triples_of_an_s(target_node=a_node,
                                                                         instantiation_property=instantiation_property):
                            yield a_triple

    def yield_classes_with_instances(self, instantiation_property=RDF_TYPE):
        """
        It yields every class URI that has at least a declared instance
        :param instantiation_property:
        :return:
        """
        raise NotImplementedError()



    def _is_an_unprefixed_iri(self, an_iri, strict_syntax_with_uri_corners=True):
        if strict_syntax_with_uri_corners:
            return an_iri[0] == "<" and an_iri[-1] == ">"
        else:
            return an_iri.startswith("http://")  # Getting kicked in the chicken nuggets is worse than this decision



from rdflib.graph import Graph, URIRef, Literal, BNode
from shexer.io.graph.yielder.base_triples_yielder import BaseTriplesYielder
from shexer.consts import N3, TURTLE, RDF_XML, NT, JSON_LD, ZIP, GZ, XZ

from shexer.model.Literal import Literal as model_Literal
from shexer.model.IRI import IRI as model_IRI
from shexer.model.bnode import BNode as model_BNode
from shexer.model.property import Property as model_Property

from shexer.utils.uri import decide_literal_type
from shexer.utils.compression import get_content_gz_file, get_content_zip_internal_file, get_content_xz_file

_SUPPORTED_FORMATS = [N3, TURTLE, RDF_XML, NT, JSON_LD]

_XML_WRONG_URI = "http://www.w3.org/XML/1998/namespace"


class RdflibTripleYielder(BaseTriplesYielder):
    def __init__(self, rdflib_graph, namespaces_dict=None):
        super().__init__()
        self._rdflib_graph = rdflib_graph
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
        self._triples_count = 0

        self._prefixes_parsed = False

    def yield_triples(self, parse_namespaces=True):
        self._reset_count()
        tmp_graph = self._get_tmp_graph()
        if parse_namespaces:
            self._integrate_namespaces_from_parsed_graph(tmp_graph, self._namespaces_dict)
            self._prefixes_parsed = True
        for sub, pred, obj in tmp_graph:
            yield (
                self._turn_rdflib_token_into_model_obj(sub),
                self._turn_rdflib_prop_into_model_obj(pred),
                self._turn_rdflib_token_into_model_obj(obj)
            )
            self._triples_count += 1

    @property
    def rdflib_graph(self):
        return self._rdflib_graph


    def _turn_rdflib_token_into_model_obj(self, rdflib_obj):
        if type(rdflib_obj) == URIRef:
            return model_IRI(str(rdflib_obj))
        elif type(rdflib_obj) == Literal:
            return self._turn_into_model_literal(rdflib_obj)
        elif type(rdflib_obj) == BNode:
            return model_BNode(identifier=str(rdflib_obj))
        else:
            raise ValueError("Not recognized type of rdflib element: " + type(rdflib_obj) + " ( " + str(rdflib_obj) + " )")

    def _turn_rdflib_prop_into_model_obj(self, rdflib_obj):
        if type(rdflib_obj) == URIRef:
            return model_Property(str(rdflib_obj))
        else:
            raise ValueError("Trying to convert into a model property en element which is not "
                             "supposed to be a property: " + type(rdflib_obj) + " ( " + str(rdflib_obj) + " )")

    def _get_tmp_graph(self):
        return self._rdflib_graph

    @staticmethod
    def _integrate_namespaces_from_parsed_graph(a_graph, namespaces_dict):

        for a_prefix_namespace_tuple in a_graph.namespaces():
            candidate_uri = str(a_prefix_namespace_tuple[1])
            if candidate_uri not in namespaces_dict:
                if candidate_uri == _XML_WRONG_URI:  # XML fix...
                    candidate_uri += "/"             # XML fix...
                namespaces_dict[candidate_uri] = str(a_prefix_namespace_tuple[0])
            # There is no else here. In case of conflict between the parsed content and the dict provided by the user,
            # the user's one have priority


    @staticmethod
    def _turn_into_model_literal(rdflib_literal):
        content = str(rdflib_literal)
        if rdflib_literal.language is not None:
            content = '"' + content + '"@' + rdflib_literal.language
        return model_Literal(content=content,
                             elem_type=str(rdflib_literal.datatype)
                             if rdflib_literal.datatype is not None
                             else decide_literal_type(content))


    @property
    def yielded_triples(self):
        return self._triples_count

    @property
    def error_triples(self):
        return 0  # With rdflib, a single error will cause to fail the parsing process


    @property
    def namespaces(self):
        if not self._prefixes_parsed:
            tmp_graph = self._get_tmp_graph()
            self._integrate_namespaces_from_parsed_graph(tmp_graph, self._namespaces_dict)
            self._prefixes_parsed = True
        return self._namespaces_dict


    def _reset_count(self):
        self._triples_count = 0


class RdflibParserTripleYielder(RdflibTripleYielder):

    def __init__(self, input_format=TURTLE, source=None, allow_untyped_numbers=False, raw_graph=None,
                 namespaces_dict=None, compression_mode=None, zip_archive_file=None):
        """

        :param input_format:
        :param source: It can be local (a file path) or remote (an url to download some content)
        :param namespaces_to_ignore:
        :param allow_untyped_numbers:
        :param raw_graph:
        :param namespaces_dict:
        """

        super().__init__(rdflib_graph=None,
                         namespaces_dict=namespaces_dict)
        self._check_input_format(input_format)
        self._input_format = input_format
        self._source = source
        self._compression_mode = compression_mode
        self._zip_archive_file = zip_archive_file
        self._allow_untyped_numbers = allow_untyped_numbers
        self._raw_graph = raw_graph
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
                                              # This object can be modified (and will be consumed externaly)
                                              # when parse_namespaces in yiled_triples() is set to True

        self._triples_count = 0

        self._prefixes_parsed = False


    def _get_tmp_graph(self):
        result = Graph()
        if self._compression_mode is not None:
            self._parse_compressed_files(result)
        elif self._source is not None:
            result.parse(source=self._source, format=self._input_format)
        else:
            result.parse(data=self._raw_graph, format=self._input_format)
        return result

    def _parse_compressed_files(self, rdflib_graph):
        if self._compression_mode == GZ:
            rdflib_graph.parse(data=get_content_gz_file(self._source), format=self._input_format)
        elif self._compression_mode == ZIP:
            rdflib_graph.parse(data=get_content_zip_internal_file(base_archive=self._zip_archive_file,
                                                                  target_file=self._source),
                               format=self._input_format)
        elif self._compression_mode == XZ:
            rdflib_graph.parse(data=get_content_xz_file(self._source), format=self._input_format)
        else:
            raise ValueError("Unknown compression format")

    @staticmethod
    def _check_input_format(input_format):
        if input_format not in _SUPPORTED_FORMATS:
            raise ValueError("Unsupported input format: " + input_format)








from shexer.io.graph.yielder.base_triples_yielder import BaseTriplesYielder
from shexer.consts import RDF_TYPE
from shexer.utils.triple_yielders import tune_token, tune_prop, tune_subj
from shexer.utils.uri import add_corners_if_needed, add_corners_if_it_is_an_uri


class SgraphFromSelectorsTripleYielder(BaseTriplesYielder):

    def __init__(self, shape_map, depth=1, classes_at_last_level=True, instantiation_property=RDF_TYPE,
                 strict_syntax_with_corners=False, allow_untyped_numbers=False, inverse_paths=False):
        super().__init__()
        self._shape_map = shape_map
        self._depth = depth
        self._classes_at_last_level = classes_at_last_level
        self._instantiation_property = instantiation_property
        self._strict_syntax_with_corners = strict_syntax_with_corners
        self._allow_untyped_numbers = allow_untyped_numbers
        self._inverse_paths = inverse_paths


    def yield_triples(self):
        target_nodes = self._collect_every_target_node()
        sgraph = self._shape_map.get_sgraph()
        for a_triple in self._yield_relevant_sgraph_triples(target_nodes, sgraph):
            yield a_triple


    def _collect_every_target_node(self):
        result = set()
        for an_item in self._shape_map.yield_items():
            for a_node in an_item.node_selector.get_target_nodes():
                result.add(a_node)
        return list(result)


    def _yield_relevant_sgraph_triples(self, target_nodes, sgraph):
        for a_triple in self._yield_relevant_direct_triples(target_nodes, sgraph):
            yield a_triple
        if self._inverse_paths:
            for a_triple in self._yield_relevant_inverse_triples(target_nodes, sgraph):
                yield a_triple

    def _yield_relevant_direct_triples(self, target_nodes, sgraph):
        for s, p, o in sgraph.yield_p_o_triples_of_target_nodes(target_nodes=target_nodes,
                                                                depth=self._depth,
                                                                classes_at_last_level=self._classes_at_last_level,
                                                                instantiation_property=self._instantiation_property,
                                                                already_visited=None,
                                                                strict_syntax_with_uri_corners=self._strict_syntax_with_corners):
            yield (tune_subj(a_token=add_corners_if_it_is_an_uri(s)),
                   tune_prop(a_token=add_corners_if_needed(p)),
                   tune_token(a_token=add_corners_if_it_is_an_uri(o),
                              allow_untyped_numbers=self._allow_untyped_numbers)
                   )

    def _yield_relevant_inverse_triples(self, target_nodes, sgraph):
        for s, p, o in sgraph.yield_s_p_triples_of_target_nodes(target_nodes=target_nodes,
                                                                depth=self._depth,
                                                                classes_at_last_level=self._classes_at_last_level,
                                                                instantiation_property=self._instantiation_property,
                                                                already_visited=None,
                                                                strict_syntax_with_uri_corners=self._strict_syntax_with_corners):
            yield (tune_subj(a_token=add_corners_if_it_is_an_uri(s)),
                   tune_prop(a_token=add_corners_if_needed(p)),
                   tune_token(a_token=add_corners_if_it_is_an_uri(o),
                              allow_untyped_numbers=self._allow_untyped_numbers)
                   )












class RawStringLineReader():

    def __init__(self, raw_string):
        self._raw_string = raw_string


    def read_lines(self):
        for a_line in self._raw_string.split("\n"):
            if a_line.strip() != "":
                yield a_line


from shexer.io.graph.yielder.base_triples_yielder import BaseTriplesYielder
from shexer.utils.triple_yielders import check_if_property_belongs_to_namespace_list

class FilterNamespacesTriplesYielder(BaseTriplesYielder):

    def __init__(self, actual_triple_yielder, namespaces_to_ignore):
        super().__init__()
        self._actual_triple_yielder = actual_triple_yielder
        self._namespaces_to_ignore = namespaces_to_ignore


    def yield_triples(self):
        for a_triple in self._actual_triple_yielder.yield_triples():
            if self._pass_filters(a_triple):
                yield a_triple

    def _pass_filters(self, a_triple):
        return not check_if_property_belongs_to_namespace_list(str_prop=str(a_triple[1]),
                                                               namespaces=self._namespaces_to_ignore)



from shexer.core.instances.abstract_instance_tracker import AbstractInstanceTracker
from shexer.utils.log import log_msg

class MixedInstanceTracker(AbstractInstanceTracker):

    def __init__(self, list_of_instance_trackers):
        self._reference_instance_tracker = list_of_instance_trackers[0]
        self._secondary_instance_trackers = [] if len(list_of_instance_trackers) <= 1 else list_of_instance_trackers[1:]
        self._disambiguator_counter = 0

    def track_instances(self, verbose=True):
        log_msg(verbose=verbose,
                msg="Starting instance tracking process with a MixedInstance tracker. "
                    "Several Instance Trackers may be launched...")
        reference_instances_dict = self._reference_instance_tracker.track_instances(verbose=verbose)
        for a_tracker in self._secondary_instance_trackers:
            self._integrate_dicts(reference_dict=reference_instances_dict,
                                  new_dict=a_tracker.track_instances(verbose=verbose),
                                  new_tracker=a_tracker)
        log_msg(verbose=verbose,
                msg="Every instance tracker has finished and their results have been integrated."
                    " {} instances have been located".format(len(reference_instances_dict)))
        return reference_instances_dict

    def _specific_disambiguator_prefix(self):
        return "mixed_"

    def _integrate_dicts(self, reference_dict, new_dict, new_tracker):
        original_classes = self._find_all_classes_in_dict(reference_dict)
        for an_instance, classes in new_dict.items():
            if an_instance not in reference_dict:
                reference_dict[an_instance] = []
            for a_class in classes:
                if a_class in original_classes:  # There is key_ambigüity, two trackers have
                                                 # identhical names for different elements
                    reference_dict[an_instance].append(self._get_label_for_ambiguous_class(a_class=a_class,
                                                                                           tracker=new_tracker))
                else:
                    reference_dict[an_instance].append(a_class)


    def _find_all_classes_in_dict(self, instances_dict):
        result = set()
        for classes in instances_dict.values():
            for a_class in classes:
                result.add(a_class)
        return result



    def _get_label_for_ambiguous_class(self, a_class, tracker):
        return tracker.disambiguator_prefix + a_class












from shexer.core.shexing.strategy.abstract_shexing_strategy import AbstractShexingStrategy
from shexer.utils.shapes import build_shapes_name_for_class_uri
from shexer.model.statement import Statement
from shexer.model.shape import Shape

_POS_FEATURES_DIRECT = 0
_POS_FEATURES_INVERSE = 1


class DirectAndInverseShexingStrategy(AbstractShexingStrategy):

    def __init__(self, class_shexer):
        super().__init__(class_shexer)
        self._class_profile_dict = self._class_shexer._class_profile_dict
        self._shapes_namespace = self._class_shexer._shapes_namespace
        self._class_counts_dict = self._class_shexer._class_counts_dict

    def remove_statements_to_gone_shapes(self, shape, shape_names_to_remove):
        shape.direct_statements = self._statements_without_shapes_to_remove(
            original_statements=shape.direct_statements,
            shape_names_to_remove=shape_names_to_remove)
        shape.inverse_statements = self._statements_without_shapes_to_remove(
            original_statements=shape.inverse_statements,
            shape_names_to_remove=shape_names_to_remove)

    def _yield_base_shapes_direction_aware(self, acceptance_threshold):
        for a_class_key in self._class_profile_dict:
            name = build_shapes_name_for_class_uri(class_uri=a_class_key,
                                                   shapes_namespace=self._shapes_namespace)
            number_of_instances = float(self._class_counts_dict[a_class_key])

            direct_statements = self._build_base_direct_statements(acceptance_threshold, a_class_key,
                                                                   number_of_instances)
            inverse_statements = self._build_base_inverse_statements(acceptance_threshold=acceptance_threshold,
                                                                     class_key=a_class_key,
                                                                     number_of_instances=number_of_instances)
            yield Shape(name=name,
                        class_uri=a_class_key,
                        statements=direct_statements + inverse_statements,
                        n_instances=int(number_of_instances))

    def set_valid_shape_constraints(self, shape):
        valid_statements = self._select_valid_statements_of_shape(shape.direct_statements)
        valid_statements += self._select_valid_statements_of_shape(shape.inverse_statements)
        self._tune_list_of_valid_statements(valid_statements=valid_statements)
        shape.statements = valid_statements

    def _build_base_inverse_statements(self, acceptance_threshold, class_key, number_of_instances):
        result = []
        for a_prop_key in self._class_profile_dict[class_key][_POS_FEATURES_INVERSE]:
            for a_type_key in self._class_profile_dict[class_key][_POS_FEATURES_INVERSE][a_prop_key]:
                for a_cardinality in self._class_profile_dict[class_key][_POS_FEATURES_INVERSE][a_prop_key][a_type_key]:
                    n_occurences = self._class_profile_dict[class_key][_POS_FEATURES_INVERSE][a_prop_key][a_type_key][a_cardinality]
                    frequency = self._compute_frequency(number_of_instances,
                                                        n_occurences)
                    if frequency >= acceptance_threshold:
                        result.append(Statement(st_property=a_prop_key,
                                                st_type=a_type_key,
                                                cardinality=a_cardinality,
                                                probability=frequency,
                                                n_occurences=n_occurences,
                                                is_inverse=True))
        return result

    def _build_base_direct_statements(self, acceptance_threshold, class_key, number_of_instances):
        result = []
        for a_prop_key in self._class_profile_dict[class_key][_POS_FEATURES_DIRECT]:
            for a_type_key in self._class_profile_dict[class_key][_POS_FEATURES_DIRECT][a_prop_key]:
                for a_cardinality in self._class_profile_dict[class_key][_POS_FEATURES_DIRECT][a_prop_key][a_type_key]:
                    n_occurences = self._class_profile_dict[class_key][_POS_FEATURES_DIRECT][a_prop_key][a_type_key][a_cardinality]
                    frequency = self._compute_frequency(number_of_instances,
                                                        n_occurences)
                    if frequency >= acceptance_threshold:
                        result.append(Statement(st_property=a_prop_key,
                                                st_type=a_type_key,
                                                cardinality=a_cardinality,
                                                probability=frequency,
                                                is_inverse=False,
                                                n_occurences=n_occurences))
        return result

    # def _set_serializer_object_for_statements(self, statement):
    #     statement.serializer_object = BaseStatementSerializer(
    #         instantiation_property_str=self._instantiation_property_str,
    #         disable_comments=self._disable_comments,
    #         is_inverse=statement.is_inverse)
    #
    # def _get_serializer_for_choice_statement(self):
    #     return FixedPropChoiceStatementSerializer(
    #         instantiation_property_str=self._instantiation_property_str,
    #         disable_comments=self._disable_comments,
    #         is_inverse=statement.is_inverse)




from shexer.io.line_reader.file_line_reader import FileLineReader
from shexer.io.line_reader.raw_string_line_reader import RawStringLineReader
from shexer.io.line_reader.gz_line_reader import GzFileLineReader
from shexer.io.line_reader.zip_file_line_reader import ZipFileLineReader
from shexer.io.line_reader.xz_line_reader import XzFileLineReader
from shexer.utils.obj_references import check_just_one_not_none
from shexer.consts import ZIP, GZ, XZ

class BaseTriplesYielder(object):

    def __init__(self):
        pass

    def _decide_line_reader(self, raw_graph, source_file,
                            compression_mode=None,
                            zip_base_archive=None):
        check_just_one_not_none((source_file, "source_file"),
                                (raw_graph, "raw_graph"))
        if raw_graph is not None:
            return RawStringLineReader(raw_string=raw_graph)
        elif compression_mode is None:
            return FileLineReader(source_file=source_file)
        elif compression_mode == GZ:
            return GzFileLineReader(gz_file=source_file)
        elif compression_mode == ZIP:
            return ZipFileLineReader(zip_archive=zip_base_archive,
                                     zip_target=source_file)
        elif compression_mode == XZ:
            return XzFileLineReader(xz_file=source_file)
        else:
            raise ValueError("Unsupported compression mode: {}".format(compression_mode))

    def yield_triples(self):
        raise NotImplementedError("Implement this method in derived classes")


from shexer.utils.obj_references import check_just_one_not_none

from shexer.consts import SHEXC, SHACL_TURTLE, NT, TSV_SPO, N3, TURTLE, TURTLE_ITER, \
    RDF_XML, FIXED_SHAPE_MAP, JSON_LD, RDF_TYPE, SHAPES_DEFAULT_NAMESPACE, ZIP, GZ, XZ, \
    ALL_EXAMPLES, CONSTRAINT_EXAMPLES, SHAPE_EXAMPLES
from shexer.utils.factories.class_profiler_factory import get_class_profiler
from shexer.utils.factories.instance_tracker_factory import get_instance_tracker
from shexer.utils.factories.class_shexer_factory import get_class_shexer
from shexer.utils.factories.remote_graph_factory import get_remote_graph_if_needed
from shexer.utils.factories.shape_map_factory import get_shape_map_if_needed
from shexer.io.profile.formater.abstract_profile_serializer import AbstractProfileSerializer
from shexer.utils.factories.shape_serializer_factory import get_shape_serializer, get_uml_serializer
from shexer.utils.namespaces import find_adequate_prefix_for_shapes_namespaces
from shexer.utils.log import log_msg
from shexer.utils.uri import unprefixize_uri_if_possible
from shexer.utils.dict import reverse_keys_and_values
from shexer.consts import RATIO_INSTANCES


class Shaper(object):

    def __init__(self, target_classes=None,
                 file_target_classes=None,
                 input_format=NT,
                 instances_file_input=None,
                 graph_file_input=None,
                 graph_list_of_files_input=None,
                 raw_graph=None,
                 url_graph_input=None,
                 rdflib_graph=None,
                 list_of_url_input=None,
                 namespaces_dict=None,
                 # namespaces_dict_file=None,
                 instantiation_property=RDF_TYPE,
                 namespaces_to_ignore=None,
                 infer_numeric_types_for_untyped_literals=True,
                 discard_useless_constraints_with_positive_closure=True,
                 all_instances_are_compliant_mode=True,
                 keep_less_specific=True,
                 all_classes_mode=False,
                 shape_map_file=None,
                 shape_map_raw=None,
                 depth_for_building_subgraph=1,
                 track_classes_for_entities_at_last_depth_level=False,
                 strict_syntax_with_corners=False,
                 url_endpoint=None,
                 shape_map_format=FIXED_SHAPE_MAP,
                 shape_qualifiers_mode=False,
                 namespaces_for_qualifier_props=None,
                 remove_empty_shapes=True,
                 disable_comments=False,
                 disable_or_statements=True,
                 allow_opt_cardinality=True,
                 disable_exact_cardinality=False,
                 shapes_namespace=SHAPES_DEFAULT_NAMESPACE,
                 limit_remote_instances=-1,
                 wikidata_annotation=False,
                 inverse_paths=False,
                 compression_mode=None,
                 decimals=-1,
                 instances_report_mode=RATIO_INSTANCES,
                 disable_endpoint_cache=False,
                 detect_minimal_iri=False,
                 allow_redundant_or=False,
                 instances_cap=-1,
                 examples_mode=None
                 ):
        """

        :param target_classes:
        :param file_target_classes:
        :param input_format:
        :param instances_file_input:
        :param graph_file_input:
        :param graph_list_of_files_input:
        :param raw_graph:
        :param url_graph_input:
        :param rdflib_graph:
        :param list_of_url_input:
        :param namespaces_dict:
        :param instantiation_property:
        :param namespaces_to_ignore:
        :param infer_numeric_types_for_untyped_literals:
        :param discard_useless_constraints_with_positive_closure:
        :param all_instances_are_compliant_mode:
        :param keep_less_specific:
        :param all_classes_mode:
        :param shape_map_file:
        :param shape_map_raw:
        :param depth_for_building_subgraph:
        :param track_classes_for_entities_at_last_depth_level:
        :param strict_syntax_with_corners:
        :param url_endpoint:
        :param shape_map_format:
        :param shape_qualifiers_mode:
        :param namespaces_for_qualifier_props:
        :param remove_empty_shapes:
        :param disable_comments:
        :param disable_or_statements:
        :param allow_opt_cardinality:
        :param disable_exact_cardinality:
        :param shapes_namespace:
        :param limit_remote_instances:
        :param wikidata_annotation:
        :param inverse_paths:
        :param compression_mode:
        :param decimals:
        :param instances_report_mode:
        :param disable_endpoint_cache:
        :param detect_minimal_iri:
        :param allow_redundant_or:
        :param instances_cap:
        :param examples_mode:
        """

        check_just_one_not_none((graph_file_input, "graph_file_input"),
                                (graph_list_of_files_input, "graph_list_of_files_input"),
                                (raw_graph, "raw_graph"),
                                (url_graph_input, "url_input"),
                                (list_of_url_input, "list_of_url_input"),
                                (url_endpoint, "url_endpoint"),
                                (rdflib_graph, "rdflib_graph")
                                )

        # check_one_or_zero_not_none((namespaces_dict, "namespaces_dict"),
        #                            (namespaces_dict_file, "namespaces_dict_file"))

        self._check_target_classes(target_classes=target_classes,
                                   file_target_classes=file_target_classes,
                                   all_classes_mode=all_classes_mode,
                                   shape_map_raw=shape_map_raw,
                                   shape_map_file=shape_map_file)

        self._check_or_config(or_disabled=disable_or_statements,
                              enable_redundant=allow_redundant_or)

        #TODO ---> Param check of shape_map and graph_via_shape_map

        self._check_input_format(input_format)

        self._check_compression_mode(compression_mode, url_endpoint, url_graph_input, list_of_url_input)

        self._check_examples_mode(examples_mode)

        self._target_classes = target_classes
        self._file_target_classes = file_target_classes
        self._input_format = input_format
        self._instances_file_input = instances_file_input
        self._graph_file_input = graph_file_input
        self._graph_list_of_files_input = graph_list_of_files_input
        self._url_graph_input = url_graph_input
        self._list_of_url_input = list_of_url_input
        self._rdflib_graph = rdflib_graph
        self._namespaces_dict = namespaces_dict if namespaces_dict is not None else {}
        self._instantiation_property = \
            unprefixize_uri_if_possible(instantiation_property,
                                        include_corners=False,
                                        prefix_namespaces_dict=reverse_keys_and_values(self._namespaces_dict))
        self._namespaces_to_ignore = namespaces_to_ignore
        self._infer_numeric_types_for_untyped_literals = infer_numeric_types_for_untyped_literals
        self._discard_useles_constraints_with_positive_closure = discard_useless_constraints_with_positive_closure
        self._all_compliant_mode = all_instances_are_compliant_mode
        self._keep_less_specific = keep_less_specific
        self._raw_graph = raw_graph
        self._all_classes_mode = all_classes_mode
        self._shape_map_file = shape_map_file
        self._shape_map_raw = shape_map_raw
        self._decimals = decimals
        self._instances_report_mode = instances_report_mode
        self._disable_endpoint_cache=disable_endpoint_cache
        self._allow_redundant_or = allow_redundant_or
        self._instances_cap = instances_cap

        self._remove_empty_shapes=remove_empty_shapes
        self._disable_comments = disable_comments
        self._disable_or_statements = disable_or_statements
        self._allow_opt_cardinality = allow_opt_cardinality
        self._disable_exact_cardinality = disable_exact_cardinality
        # TODO: REMOVE THE _limit_remote_instances PARAMETER IN FUTURE RELEASES
        self._limit_remote_instances = limit_remote_instances if instances_cap==-1 else instances_cap
        self._wikidata_annotation = wikidata_annotation
        self._inverse_paths = inverse_paths
        self._detect_minimal_iri = detect_minimal_iri
        self._examples_mode = examples_mode

        self._compression_mode = compression_mode

        self._depth_for_building_subgraph = depth_for_building_subgraph
        self._track_classes_for_entities_at_last_depth_level = track_classes_for_entities_at_last_depth_level
        self._url_endpoint = url_endpoint
        self._strict_syntax_with_corners = strict_syntax_with_corners
        self._shape_map_format = shape_map_format
        self._shape_qualifiers_mode = shape_qualifiers_mode
        self._namespaces_for_qualifier_props = namespaces_for_qualifier_props
        self._shapes_namespace = shapes_namespace

        self._add_shapes_namespaces_to_namespaces_dict()


        #The following two atts are used for optimizations
        self._built_remote_graph = get_remote_graph_if_needed(endpoint_url=url_endpoint,
                                                              store_locally=not disable_endpoint_cache)
        self._built_shape_map = get_shape_map_if_needed(sm_format=self._shape_map_format,
                                                        remote_sgraph=self._built_remote_graph,
                                                        namespaces_prefix_dict=self._namespaces_dict,
                                                        target_classes=self._target_classes,
                                                        file_target_classes=self._file_target_classes,
                                                        shape_map_file=self._shape_map_file,
                                                        shape_map_raw=self._shape_map_raw,
                                                        instantiation_property=self._instantiation_property,
                                                        shape_map_already_built=None,
                                                        rdflib_graph=self._rdflib_graph,
                                                        raw_graph=self._raw_graph,
                                                        input_format=self._input_format,
                                                        source_file_graph=self._graph_file_input,
                                                        limit_remote_instances=self._limit_remote_instances)



        self._instance_tracker = None
        self._target_classes_dict = None
        self._class_profiler = None
        self._profile = None
        self._class_counts = None
        self._class_min_iris = None
        self._class_shexer = None
        self._shape_list = None

    def profile_graph(self, string_output=False, output_file=None, verbose=False):
        self._check_correct_output_params(string_output, output_file)
        if self._target_classes_dict is None:
            self._launch_instance_tracker(verbose=verbose)
        if self._profile is None:
            self._launch_class_profiler(verbose=verbose)
        log_msg(verbose=verbose,
                msg="Building_output...")
        if string_output:
            return AbstractProfileSerializer(self._profile).get_string_representation()
        return AbstractProfileSerializer(self._profile).write_profile_to_file(target_file=output_file)

    def shex_graph(self, string_output=False,
                   output_file=None,
                   output_format=SHEXC,
                   acceptance_threshold=0,
                   verbose=False,
                   to_uml_path=None):
        """
        :param string_output:
        :param output_file:
        :param output_format:
        :param acceptance_threshold:
        :param verbose:
        :param to_uml_path:
        :return:
        """
        self._check_correct_output_params(string_output, output_file, to_uml_path)
        self._check_output_format(output_format)
        self._check_aceptance_threshold(acceptance_threshold)
        if self._target_classes_dict is None:
            self._launch_instance_tracker(verbose=verbose)
        if self._profile is None:
            self._launch_class_profiler(verbose=verbose)
        if self._shape_list is None:
            self._launch_class_shexer(acceptance_threshold=acceptance_threshold,
                                      verbose=verbose)
        log_msg(verbose=verbose,
                msg="Building_output...")

        if to_uml_path is not None:
            log_msg(verbose=verbose,
                    msg="Trying to generat UML diagram...")
            try:
                self._generate_uml_diagram(to_uml_path)
                log_msg(verbose=verbose,
                        msg="UML diagram generated...")
            except ResourceWarning as e:  # I think this is related to UMLPlant and I can't close the connection from here
                pass

        if string_output or output_file is not None:
            log_msg(verbose=verbose,
                    msg="Generating text serialization...")
            serializer = self._build_shapes_serializer(target_file=output_file,
                                                       string_return=string_output,
                                                       output_format=output_format)

            return serializer.serialize_shapes()  # If string return is active, returns string.


    def _generate_uml_diagram(self, to_uml_path):
        serializer = get_uml_serializer(shapes_list=self._shape_list,
                                        image_path=to_uml_path,
                                        namespaces_dict=self._namespaces_dict)
        serializer.serialize_shapes()




    def _add_shapes_namespaces_to_namespaces_dict(self):
        self._namespaces_dict[self._shapes_namespace] = \
            find_adequate_prefix_for_shapes_namespaces(self._namespaces_dict)

    def _launch_class_profiler(self, verbose=False):
        if self._class_profiler is None:
            self._class_profiler = self._build_class_profiler()
        self._profile, self._class_counts, self._class_min_iris = self._class_profiler.profile_classes(verbose=verbose)

    def _launch_class_shexer(self, acceptance_threshold, verbose=False):
        if self._class_shexer is None:
            self._class_shexer = self._build_class_shexer()
        self._shape_list = self._class_shexer.shex_classes(acceptance_threshold=acceptance_threshold,
                                                           verbose=verbose)

    def _launch_instance_tracker(self, verbose=False):
        if self._instance_tracker is None:
            self._instance_tracker = self._build_instance_tracker()
        self._target_classes_dict = self._instance_tracker.track_instances(verbose=verbose)

    def _build_class_shexer(self):
        return get_class_shexer(class_counts=self._class_counts,
                                class_profile_dict=self._profile,
                                original_target_classes=self._target_classes,
                                original_shape_map=self._built_shape_map,
                                remove_empty_shapes=self._remove_empty_shapes,
                                discard_useless_constraints_with_positive_closure=
                                self._discard_useles_constraints_with_positive_closure,
                                keep_less_specific=self._keep_less_specific,
                                all_compliant_mode=self._all_compliant_mode,
                                instantiation_property=self._instantiation_property,
                                disable_or_statements=self._disable_or_statements,
                                disable_comments=self._disable_comments,
                                namespaces_dict=self._namespaces_dict,
                                allow_opt_cardinality=self._allow_opt_cardinality,
                                disable_exact_cardinality=self._disable_exact_cardinality,
                                shapes_namespace=self._shapes_namespace,
                                inverse_paths=self._inverse_paths,
                                decimals=self._decimals,
                                instances_report_mode=self._instances_report_mode,
                                detect_minimal_iri=self._detect_minimal_iri,
                                class_min_iris=self._class_min_iris,
                                allow_redundant_or=self._allow_redundant_or,

                                )

    def _build_shapes_serializer(self, target_file, string_return, output_format):
        return get_shape_serializer(shapes_list=self._shape_list,
                                    target_file=target_file,
                                    string_return=string_return,
                                    namespaces_dict=self._namespaces_dict,
                                    output_format=output_format,
                                    instantiation_property=self._instantiation_property,
                                    disable_comments=self._disable_comments,
                                    wikidata_annotation=self._wikidata_annotation,
                                    instances_report_mode=self._instances_report_mode,
                                    detect_minimal_iri=self._detect_minimal_iri,
                                    shape_features_examples=self._class_min_iris,
                                    examples_mode=self._examples_mode,
                                    inverse_paths=self._inverse_paths)

    def _build_class_profiler(self):
        return get_class_profiler(target_classes_dict=self._target_classes_dict,
                                  source_file=self._graph_file_input,
                                  list_of_source_files=self._graph_list_of_files_input,
                                  input_format=self._input_format,
                                  instantiation_property_str=self._instantiation_property,
                                  namespaces_to_ignore=self._namespaces_to_ignore,
                                  infer_numeric_types_for_untyped_literals=self._infer_numeric_types_for_untyped_literals,
                                  raw_graph=self._raw_graph,
                                  namespaces_dict=self._namespaces_dict,
                                  url_input=self._url_graph_input,
                                  list_of_url_input=self._list_of_url_input,
                                  rdflib_graph=self._rdflib_graph,
                                  shape_map_file=self._shape_map_file,
                                  shape_map_raw=self._shape_map_raw,
                                  track_classes_for_entities_at_last_depth_level=self._track_classes_for_entities_at_last_depth_level,
                                  depth_for_building_subgraph=self._depth_for_building_subgraph,
                                  url_endpoint=self._url_endpoint,
                                  strict_syntax_with_corners=self._strict_syntax_with_corners,
                                  target_classes=self._target_classes,
                                  file_target_classes=self._file_target_classes,
                                  built_remote_graph=self._built_remote_graph,
                                  built_shape_map=self._built_shape_map,
                                  remove_empty_shapes=self._remove_empty_shapes,
                                  limit_remote_instances=self._limit_remote_instances,
                                  inverse_paths=self._inverse_paths,
                                  all_classes_mode=self._all_classes_mode,
                                  compression_mode=self._compression_mode,
                                  disable_endpoint_cache=self._disable_endpoint_cache,
                                  detect_minimal_iri=self._detect_minimal_iri,
                                  examples_mode=self._examples_mode)


    def _build_instance_tracker(self):
        return get_instance_tracker(instances_file_input=self._instances_file_input,
                                    graph_file_input=self._graph_file_input,
                                    graph_list_of_files_input=self._graph_list_of_files_input,
                                    target_classes=self._target_classes,
                                    file_target_classes=self._file_target_classes,
                                    input_format=self._input_format,
                                    instantiation_property=self._instantiation_property,
                                    infer_numeric_types_for_untyped_literals=self._infer_numeric_types_for_untyped_literals,
                                    raw_graph=self._raw_graph,
                                    all_classes_mode=self._all_classes_mode,
                                    namespaces_dict=self._namespaces_dict,
                                    url_input=self._url_graph_input,
                                    list_of_url_input=self._list_of_url_input,
                                    rdflib_graph=self._rdflib_graph,
                                    shape_map_file=self._shape_map_file,
                                    shape_map_raw=self._shape_map_raw,
                                    track_classes_for_entities_at_last_depth_level=self._track_classes_for_entities_at_last_depth_level,
                                    depth_for_building_subgraph=self._depth_for_building_subgraph,
                                    url_endpoint=self._url_endpoint,
                                    strict_syntax_with_corners=self._strict_syntax_with_corners,
                                    shape_map_format=self._shape_map_format,
                                    namespaces_for_qualifier_props=self._namespaces_for_qualifier_props,
                                    shape_qualifiers_mode=self._shape_qualifiers_mode,
                                    built_remote_graph=self._built_remote_graph,
                                    built_shape_map=self._built_shape_map,
                                    shapes_namespace=self._shapes_namespace,
                                    limit_remote_instances=self._limit_remote_instances,
                                    inverse_paths=self._inverse_paths,
                                    compression_mode=self._compression_mode,
                                    disable_endpoint_cache=self._disable_endpoint_cache,
                                    instances_cap=self._instances_cap)


    @staticmethod
    def _check_correct_output_params(string_output, target_file, to_uml_path):
        if not string_output and target_file is None and to_uml_path is None:
            raise ValueError("You must provide a target path , set string output to True and/or give a value to to_uml_path")

    @staticmethod
    def _check_input_format(input_format):
        if input_format not in [NT, TSV_SPO, N3, TURTLE, RDF_XML, JSON_LD, TURTLE_ITER]:
            raise ValueError("Currently unsupported input format: " + input_format)

    @staticmethod
    def _check_compression_mode(compression_mode, url_endpoint, url_graph_input, list_of_url_input):
        if compression_mode not in [ZIP, GZ, XZ, None]:
            raise ValueError("Unknownk compression mode: {}. "
                             "The currently supported compression formats are {}.".format(
                compression_mode,
                ", ".join([ZIP, GZ, XZ])))
        if compression_mode is not None and (url_endpoint is not None or url_graph_input is not None or list_of_url_input is not None):
            raise ValueError("You've chosed some compression mode ({}) to work with remote sources."
                             "Currently, sheXer can only parse compressed local files".format(compression_mode))


    @staticmethod
    def _check_target_classes(target_classes, file_target_classes, all_classes_mode, shape_map_file, shape_map_raw):
        if not all_classes_mode:
            check_just_one_not_none((target_classes, "target_classes"),
                                    (file_target_classes, "file_target_classes"),
                                    (shape_map_file, "shape_map_file"),
                                    (shape_map_raw, "shape_map_raw")
                                    )
        else:
            if target_classes is not None or file_target_classes is not None:
                raise ValueError("You must provide a list of target classes XOR set all_classes_mode to True")
            # But all_classes mode is compatible with shape_map_selectors. Setting all_classes_mode = True and
            # providing some selectros will cause shexer to shex both the shapes specified in the selectors
            # and to create a shape for eahc element with an instance in the target graph

    @staticmethod
    def _check_output_format(output_format):
        if output_format not in [SHEXC, SHACL_TURTLE]:
            raise ValueError("Currently unsupported output format: " + output_format)

    @staticmethod
    def _check_or_config(or_disabled, enable_redundant):
        if or_disabled and enable_redundant:
            raise ValueError("You are indicating that you'd like to have disjunction constraints including the macro "
                             "IRI, but also that you do not want to have or constraints. Please, check your configuration"
                             "of the disable_or_statements and allow_redundant_or paremeters")

    @staticmethod
    def _check_aceptance_threshold(aceptance_threshold):
        if aceptance_threshold < 0 or aceptance_threshold > 1:
            raise ValueError("The acceptance threshold must be a value in [0,1]")

    @staticmethod
    def _check_examples_mode(examples_mode):
        if examples_mode not in [None, ALL_EXAMPLES, CONSTRAINT_EXAMPLES, SHAPE_EXAMPLES]:
            raise ValueError("The examples mode param should be set to None or using one of the values in shexer.const, section \"EXAMPLES\" ")



from gzip import open as gzopen
from zipfile import ZipFile
from xz import open as xzopen


def get_content_xz_file(xz_path):
    with xzopen(xz_path, "r") as in_stream:
        return in_stream.read()


def get_content_gz_file(gz_path):
    with gzopen(gz_path, "r") as in_stream:
        return in_stream.read()


def yield_contents_zip_dir(zip_path):
    with ZipFile(zip_path, 'r') as zip:
        for a_file_path in zip.filelist:
            with zip.open(a_file_path) as in_file:
                yield in_file.read()

def get_content_zip_internal_file(base_archive, target_file):
    with base_archive.open(target_file, "r") as in_file:
        return in_file.read()

def list_of_zip_internal_files(zip_base_archive):
    return zip_base_archive.namelist()


from shexer.core.instances.annotators.strategy_mode.base_strategy_mode import BaseStrategyMode
from shexer.utils.triple_yielders import check_if_property_belongs_to_namespace_list
from shexer.utils.shapes import build_shape_name_for_qualifier_prop_uri
from shexer.core.instances.pconsts import _P, _O

class ShapeQualifiersMode(BaseStrategyMode):

    def __init__(self, annotator_ref, namespaces_for_qualifiers_props, shapes_namespace):
        super().__init__(annotator_ref)
        self._namespaces_for_qualifiers_props = namespaces_for_qualifiers_props
        self._dict_of_qualifier_properties = {}
        self._shapes_namespace = shapes_namespace


    def is_relevant_triple(self, a_triple):
        if check_if_property_belongs_to_namespace_list(str_prop=a_triple[_P].iri,
                                                       namespaces=self._namespaces_for_qualifiers_props):
            return True
        return False


    def annotate_triple(self, a_triple):
        self._annotate_qualifier_shape(a_triple[_P])
        self._annotate_instance_of_a_qualifier(a_triple)


    def _annotate_instance_of_a_qualifier(self, a_triple):
        if a_triple[_O].iri not in self._instances_dict:
            self._instances_dict[a_triple[_O].iri] = []
        self._instances_dict[a_triple[_O].iri].append(self._dict_of_qualifier_properties[a_triple[_P].iri])


    def _annotate_qualifier_shape(self, a_property):
        str_prop = a_property.iri
        if str_prop not in self._dict_of_qualifier_properties:
            self._dict_of_qualifier_properties[str_prop] = \
                build_shape_name_for_qualifier_prop_uri(prop_uri=str_prop,
                                                        shapes_namespace=self._shapes_namespace)
